@@ -122,6 +122,7 @@ class TmxPicoAio:
         self.report_dispatch.update(
             {PrivateConstants.SONAR_DISTANCE: self._sonar_distance_report})
         self.report_dispatch.update({PrivateConstants.DHT_REPORT: self._dht_report})
+        self.report_dispatch.update({PrivateConstants.SPI_REPORT: self._spi_report})
 
         # up to 16 pwm pins may be simultaneously active
         self.pwm_active_count = 0
@@ -137,6 +138,13 @@ class TmxPicoAio:
         self.i2c_0_active = False
         self.i2c_1_active = False
 
+        # spi
+        self.spi_callback = None
+        self.spi_callback2 = None
+
+        self.spi_0_active = False
+        self.spi_1_active = False
+
         # the trigger pin will be the key to retrieve
         # the callback for a specific HC-SR04
         self.sonar_callbacks = {}
@@ -934,9 +942,143 @@ class TmxPicoAio:
         """
 
         await self._set_pin_mode(pin_number, PrivateConstants.AT_SERVO, min_pulse,
-                               max_pulse)
+                                 max_pulse)
         self.pico_pins[pin_number] = PrivateConstants.AT_SERVO
 
+    async def set_pin_mode_spi(self, spi_port=0, miso=16, mosi=19, clock_pin=18,
+                               clk_frequency=500000, chip_select_list=None,
+                               qualify_pins=True):
+        """
+        Specify the SPI port, SPI pins, clock frequency and an optional
+        list of chip select pins. The SPI port is configured as a "master".
+
+        :param spi_port: 0 = spi0, 1 = spi1
+
+        :param miso: SPI data receive pin
+
+        :param mosi: SPI data transmit pin (19 for
+
+        :param clock_pin: clock pin
+
+        :param clk_frequency: clock frequency in Hz.
+
+        :param chip_select_list: this is a list of pins to be used for chip select.
+                           The pins will be configured as output, and set to high
+                           ready to be used for chip select.
+                           NOTE: You must specify the chips select pins here!
+
+        :param qualify_pins: If true validate
+
+                            for spi0:
+                                 MOSI=19
+
+                                 MISO=16
+
+                                 CLOCK=18
+
+                             for spi1:
+
+                                 MOSI=15
+
+                                 MISO=12
+
+                                 CLOCK=14
+
+        command message: [command, spi port, mosi, miso, clock, freq msb,
+                          freq 3, freq 2, freq 1, number of cs pins, cs pins...]
+        """
+        # determine if the spi port is specified correctly
+        if spi_port not in [0, 1]:
+            if self.shutdown_on_exception:
+                await self.shutdown()
+            raise RuntimeError('spi port must be either a 0 or 1')
+
+        # determine if the spi gpio's are valid if qualify_pin is True.
+        if qualify_pins:
+            if spi_port == 0:
+                if mosi != 19:
+                    if self.shutdown_on_exception:
+                        await self.shutdown()
+                    raise RuntimeError('For spi0 mosi must be 19.')
+                if miso != 16:
+                    if self.shutdown_on_exception:
+                        await self.shutdown()
+                    raise RuntimeError('For spi0 miso must be 16.')
+                if clock_pin != 18:
+                    if self.shutdown_on_exception:
+                        await self.shutdown()
+                    raise RuntimeError('For spi0 clock must be 18.')
+            else:
+                if mosi != 15:
+                    if self.shutdown_on_exception:
+                        await self.shutdown()
+                    raise RuntimeError('For spi1 mosi must be 15.')
+                if miso != 12:
+                    if self.shutdown_on_exception:
+                        await self.shutdown()
+                    raise RuntimeError('For spi1 miso must be 12.')
+                if clock_pin != 14:
+                    if self.shutdown_on_exception:
+                        await self.shutdown()
+                    raise RuntimeError('For spi0 clock must be 14.')
+
+        # check if mosi, miso or clock pins have already been assigned
+        if self.pico_pins[mosi] != PrivateConstants.AT_MODE_NOT_SET:
+            if self.shutdown_on_exception:
+                await self.shutdown()
+            raise RuntimeError('MOSI pin currently in use')
+        if self.pico_pins[miso] != PrivateConstants.AT_MODE_NOT_SET:
+            if self.shutdown_on_exception:
+                await self.shutdown()
+            raise RuntimeError('MISO pin currently in use')
+        if self.pico_pins[clock_pin] != PrivateConstants.AT_MODE_NOT_SET:
+            if self.shutdown_on_exception:
+                await self.shutdown()
+            raise RuntimeError('Clock Pin pin currently in use')
+
+        if type(chip_select_list) != list:
+            if self.shutdown_on_exception:
+                await self.shutdown()
+            raise RuntimeError('chip_select_list must be in the form of a list')
+        if not chip_select_list:
+            if self.shutdown_on_exception:
+                await  self.shutdown()
+            raise RuntimeError('Chip select pins were not specified')
+        # validate chip select pins
+        for pin in chip_select_list:
+            if self.pico_pins[pin] != PrivateConstants.AT_MODE_NOT_SET:
+                if self.shutdown_on_exception:
+                    await self.shutdown()
+                raise RuntimeError(f'SPI Chip select pin {pin} is already in use!')
+
+        # test for spi port 0
+        if not spi_port:
+            self.spi_0_active = True
+        # port 1
+        else:
+            self.spi_1_active = True
+
+        # freq_msb = clk_frequency >> 8
+        # freq_lsb = clk_frequency & 0x00ff
+        freq_bytes = clk_frequency.to_bytes(4, byteorder='big')
+
+        self.pico_pins[mosi] = PrivateConstants.AT_SPI
+        self.pico_pins[miso] = PrivateConstants.AT_SPI
+        self.pico_pins[clock_pin] = PrivateConstants.AT_SPI
+
+        command = [PrivateConstants.SPI_INIT, spi_port, mosi, miso, clock_pin]
+
+        for i in range(len(freq_bytes)):
+            command.append(freq_bytes[i])
+
+        command.append(len(chip_select_list))
+
+        for pin in chip_select_list:
+            command.append(pin)
+            self.pico_pins[pin] = PrivateConstants.AT_SPI
+
+        await self._send_command(command)
+
     async def servo_write(self, pin_number, value):
         """
         Write the value to the specified servo
@@ -962,6 +1104,132 @@ class TmxPicoAio:
         # use a raw pwm write from the calculated values
         await self.pwm_write(pin_number, duty_cycle, True)
 
+    async def spi_cs_control(self, chip_select_pin, select):
+        """
+        Control an SPI chip select line
+        :param chip_select_pin: pin connected to CS
+
+        :param select: 0=select, 1=deselect
+        """
+
+        if self.pico_pins[chip_select_pin] != PrivateConstants.AT_SPI:
+            if self.shutdown_on_exception:
+                await self.shutdown()
+            raise RuntimeError(f'spi_read_blocking: Invalid chip select pin'
+                               f' {chip_select_pin}.')
+        command = [PrivateConstants.SPI_CS_CONTROL, chip_select_pin, select]
+        await self._send_command(command)
+
+    async def spi_read_blocking(self, number_of_bytes, spi_port=0, call_back=None,
+                                repeated_tx_data=0):
+        """
+        Read the specified number of bytes from the specified SPI port and
+        call the callback function with the reported data.
+
+        :param number_of_bytes: Number of bytes to read
+
+        :param spi_port: SPI port 0 or 1
+
+        :param call_back: Required callback function to report spi data as a
+                   result of read command
+
+        :param repeated_tx_data: repeated data to send
+
+        callback returns a data list:
+        [SPI_READ_REPORT, spi_port, count of data bytes, data bytes, time-stamp]
+
+        SPI_READ_REPORT = 13
+
+
+        """
+
+        if not call_back:
+            if self.shutdown_on_exception:
+                await self.shutdown()
+            raise RuntimeError('spi_read_blocking: A Callback must be specified')
+
+        if not spi_port:
+            if not self.spi_0_active:
+                if self.shutdown_on_exception:
+                    await self.shutdown()
+                raise RuntimeError(
+                    'spi_read_blocking: set_pin_mode_spi never called for spi port 0.')
+
+        elif spi_port:
+            if not self.spi_1_active:
+                if self.shutdown_on_exception:
+                    await self.shutdown()
+                raise RuntimeError(
+                    'spi_read_blocking: set_pin_mode_spi never called for spi port 1.')
+        if spi_port == 0:
+            self.spi_callback = call_back
+        else:
+            self.spi_callback2 = call_back
+
+        command = [PrivateConstants.SPI_READ_BLOCKING, spi_port, number_of_bytes,
+                   repeated_tx_data]
+        await self._send_command(command)
+
+    async def spi_set_format(self, spi_port=0, data_bits=8, spi_polarity=0, spi_phase=0):
+        """
+        Configure how the SPI serializes and de-serializes data on the wire.
+
+        :param spi_port: SPI port 0 or 1
+
+        :param data_bits: Number of data bits per transfer. Valid range = 4-16
+
+        :param spi_polarity: clock polarity. 0 or 1.
+
+        :param spi_phase: clock phase. 0 or 1.
+        """
+        if not spi_port:
+            if not self.spi_0_active:
+                if self.shutdown_on_exception:
+                    await self.shutdown()
+                raise RuntimeError(
+                    'spi_set_format: set_pin_mode_spi never called for spi port 0.')
+
+        elif spi_port:
+            if not self.spi_1_active:
+                if self.shutdown_on_exception:
+                    await self.shutdown()
+                raise RuntimeError(
+                    'spi_set_format: set_pin_mode_spi never called for spi port 1.')
+        command = [PrivateConstants.SPI_SET_FORMAT, spi_port, data_bits,
+                   spi_polarity, spi_phase]
+        await self._send_command(command)
+
+    async def spi_write_blocking(self, bytes_to_write, spi_port=0):
+        """
+        Write a list of bytes to the SPI device.
+
+        :param bytes_to_write: A list of bytes to write. This must be in the form of a
+        list.
+
+        :param spi_port: SPI port 0 or 1
+
+        """
+        if not spi_port:
+            if not self.spi_0_active:
+                if self.shutdown_on_exception:
+                    await self.shutdown()
+                raise RuntimeError(
+                    'spi_write_blocking: set_pin_mode_spi never called for spi port 0.')
+
+        elif spi_port:
+            if not self.spi_1_active:
+                if self.shutdown_on_exception:
+                    await self.shutdown()
+                raise RuntimeError(
+                    'spi_write_blocking: set_pin_mode_spi never called for spi port 1.')
+        command = [PrivateConstants.SPI_WRITE_BLOCKING, spi_port,
+                   len(bytes_to_write)]
+
+        for data in bytes_to_write:
+            command.append(data)
+
+        await self._send_command(command)
+
     async def set_pin_mode_sonar(self, trigger_pin, echo_pin, callback=None):
         """
         :param trigger_pin:  Sensor trigger gpio pin
@@ -1118,6 +1386,23 @@ class TmxPicoAio:
         except (RuntimeError, SerialException):
             pass
 
+    async def _spi_report(self, report):
+        """
+        Execute callback for spi reads.
+
+        :param report: [spi_port, number of bytes read, data]
+
+        """
+
+        cb_list = [PrivateConstants.SPI_REPORT, report[0], report[1]] + report[2:]
+
+        cb_list.append(time.time())
+
+        if cb_list[1]:
+            await self.spi_callback2(cb_list)
+        else:
+            await self.spi_callback(cb_list)
+
     '''
     report message handlers
     '''
